{
  Original Name: Twitter.pas
  Copyright: Copyright (c) 2010, Simon J Stuart, All Rights Reserved
  License: Licensed under the MIT License.
  Author: Simon J Stuart
  Author's Website: http://www.lakraven.com
  Description:
    This unit contains the TTwitter component, an easy-to-use
    layer which performs all interactions with the Twitter API
    for you, leaving you with standardized Delphi classes that
    contain the required information in a friendly format.
  Author's Notes:
    You are encouraged to include a credit somewhere within any
    application(s) using this unit (and those related) resembling
    the following:
    "This application uses TTwitter, produced by Simon J Stuart
    whose website can be found at http://www.lakraven.com"
    Whilst your use of this unit (and those related) is not
    contingent on provision of a notice or credit, it is appreciated.
    I also appreicate "success stories" or comments related to my
    code. You can find details of how to contact me at my website
    (see "Author's Website" for URL).
  Acknowledgements:
    - Jim McKeeth       : Testing, code proofing, XE trials, useful suggestions,
                          being an all-round good guy!
                          (www.delphi.org)
    - Chuck Beasley     : Providing invaluable insight into OAuth with Twitter,
                          all in the necessary Delphi context.
                          His work provided a solid foundation on which I was
                          able to build... even though he didn't know it!
                          (if you're reading this, Chuck, thanks my new friend!)
                          (www.chuckbeasley.com)
    - Twitter API Staff : Answering many questions, even the stupid ones.
                          (www.twitter.com)

  Version: 0.1 OFFICIAL OPEN BETA RELEASE
  --------------------------------------------------------------------------------
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
  --------------------------------------------------------------------------------
  CODE NOTES:
  - Should you encounter any errors, send details as a comment to:
    http://www.lakraven.com/delphi-stuff/ttwitter
    Please include the code which causes the error, as well as any Message(s)
    you see at the point of error. You can also contact me on Twitter: @LaKraven
  - Eventually a lot of Twitter actions will become Threaded to provide a smooth
    runtime experience in your application(s) for the user(s).
    At present everything executes in the same Thread (easier debugging)

  IMPORTANT NOTICE: You are currently looking at an Open BETA version of TTwitter!
                    This code is subject to a LOT of changes, and whilst you may
                    find portions of this code at present to be messy... rest
                    assured that upon final release, it'll all be nice and clean!
}

unit Twitter;

interface

{$I Twitter.inc} // TTwitter's Configuration Data

uses
  Windows, Classes, SysUtils, Generics.Collections,
  Twitter_Common, Twitter_StringComposer, Twitter_OAuth, Twitter_OLogin,
  Twitter_Data, Twitter_TokenStore, Twitter_HTTP;

type
  {
    Forward Declaration of Classes
  }
  TTwitter = class;
  TTwitterUserManager = class;
  TTwitterTimelineManager = class;

  {
    Various "OnEvent" method types
  }
  TTwitterStatusEvent = procedure(const AResponseCode: Integer; const AMsg: String; const AAction: TTwitterStatusAction) of object;
  TTwitterExceptionEvent = procedure(const AMsg: String) of object; // For when TTwitter fails
  TTwitterUsernameChangeEvent = procedure(const ANewUsername: String) of object;
  TTwitterRawResponseEvent = procedure(const AURL, AResponse: String) of object;
  TTwitterRateLimitStatusEvent = procedure(const ARateLimitDetails: TTwitterRateLimitData) of object;

  {
    TTwitter object does all the magic! This is the actual component which registers on your
    Component Pallet.
    - All communications with Twitter are neatly exposed as native Delphi methods.
    - "Get" responses are provided to use by On<event> methods!
    - "Post" responses are returned as Function Results!
  }
  TTwitter = class(TComponent)
  private
    FApplicationDetails: TTwitterApplicationDetails;
    FAutoRateCheck: Boolean;
    FHTTP: TTwitterHTTP;
    FLoggedIn: Boolean;
    FProxy: TTwitterProxyDetails;
    FTimelines: TTwitterTimelineManager;
    FTokenManager: TTwitterTokenManager;
    FUsername: string;
    FUsers: TTwitterUserManager;
    FUserToken: TTwitterUserToken;
    FOnAsyncTimeline: TTwitterAsyncTimelineEvent;
    FOnError: TTwitterStatusEvent;
    FOnException: TTwitterExceptionEvent;
    FOnRateLimitStatus: TTwitterRateLimitStatusEvent;
    FOnRawResponse: TTwitterRawResponseEvent; // DESIGNED ONLY TO BE USED FOR DEBUGGING PURPOSES DURING TTWITTER'S OWN DEVELOPMENT!
    FOnUsernameChange: TTwitterUsernameChangeEvent;
    function GetDefaultMessage(const AResponseCode: Integer): String;
    function TwitterGET(const ARequiresAuth: Boolean; var AURL: String): String;
    function TwitterPOST(var AURL: String; const AParams: TTwitterParamFactory): String;
//    procedure AsyncTwitterGET(const ARequiresAuth: Boolean; var AURL: String; const ABetterWithAuth: Boolean = True);
//    procedure AsyncTwitterPOST(var AURL: String; const AParams: TTwitterParamFactory);
    procedure DoStatus(const AResponseCode: Integer; const AMsg: String;
      const AAction: TTwitterStatusAction; const AStatusType: TTwitterStatusType);
    function GetUserInfo(const AUsername: String; const AIncludeEntities: Boolean = True): TTwitterUser; overload;
    function GetUserInfo(const AID: Int64; const AIncludeEntities: Boolean = True): TTwitterUser; overload;
  public
    constructor Create; reintroduce; overload;
    constructor Create(AOwner: TComponent); overload; override;
    destructor Destroy; override;
    function Login: Boolean;
    procedure ForgetUser; // will remove any Stored Token/Secret for the current/named User
    procedure GetRateLimitStatus;
    procedure Logout;
    function PostTweet(const AMsg: string; const AReplyToStatusID: Int64 = 0;
      const AImage: string = ''; const ALat: Extended = 181; const ALong: Extended = 181;
      const APlaceID: string = ''; const ADisplayCoordinates: Boolean = False;
      ATrimUser: Boolean = False; AIncludeEntities: Boolean = True): TTwitterTweet;
    function Request(const AMethod: THttpMethod; var AURL: string;
      const AParams: TTwitterParamFactory; const ARequiresAuth: Boolean = True): string; overload;
    function Request<T: record>(const AMethod: THttpMethod; var AURL: string;
      const AParams: T; const ARequiresAuth: Boolean = True): string; overload;
//    procedure AsyncRequest(const AMethod: THttpMethod; var AURL: string;
//      const AParams: TTwitterParamFactory; const ARequiresAuth: Boolean = True);
  published
    property ApplicationDetails: TTwitterApplicationDetails read FApplicationDetails write FApplicationDetails;
    property AutoRateCheck: Boolean read FAutoRateCheck write FAutoRateCheck default False;
    property LoggedIn: Boolean read FLoggedIn;
    property Proxy: TTwitterProxyDetails read FProxy write FProxy;
    property Timelines: TTwitterTimelineManager read FTimelines write FTimelines;
    property TokenStore: TTwitterTokenManager read FTokenManager write FTokenManager;
    property Username: String read FUsername write FUsername;
    property Users: TTwitterUserManager read FUsers write FUsers;
    { OnEvent Methods }
    property OnAsyncTimeline: TTwitterAsyncTimelineEvent read FOnAsyncTimeline write FOnAsyncTimeline;
    property OnError: TTwitterStatusEvent read FOnError write FOnError;
    property OnException: TTwitterExceptionEvent read FOnException write FOnException;
    property OnRateLimitStatus: TTwitterRateLimitStatusEvent read FOnRateLimitStatus write FOnRateLimitStatus;
    property OnRawResponse: TTwitterRawResponseEvent read FOnRawResponse write FOnRawResponse;
    property OnUsernameChange: TTwitterUsernameChangeEvent read FOnUsernameChange write FOnUsernameChange;
  end;

  {
    TTwitterUserManager object provides us a neat means of managing Twitter Users (details and actions).
    It can cache retrieved User Details to make future interrogation of those details quicker!
  }
  TTwitterUserManager = class(TPersistent)
  private
    FTwitter: TTwitter;
    FUsers: TTwitterUsers;
    FCache,
    FIncludeEntities: Boolean;
    function GetUser(const AUsername: String): TTwitterUser; // Get user by SCREEN NAME
    function GetUserByID(const AID: Int64): TTwitterUser; // Get user by ID
  public
    constructor Create(ATwitter: TTwitter);
    destructor Destroy; override;
    property User[const AUsername: String]: TTwitterUser read GetUser; default;
    property User[const AID: Int64]: TTwitterUser read GetUserByID; default;
  published
    property Cache: Boolean read FCache write FCache default True;
    property IncludeEntities: Boolean read FIncludeEntities write FIncludeEntities default True;
  end;

  {
    TTwitterTimeline object provides us with multiple Tweets, as well as operations we can
    perform on them as appropriate.
  }
  TTwitterTimeline = class(TPersistent)
  private
    FTimelineManager: TTwitterTimelineManager;
    FTimelineType: TTwitterTimelineType;
    FCache: Boolean;
    FMaxStore: Integer;
    FTweets: TTweetArray;
    FTotal: Integer;
    FParams: TTwitterTimelineParams;
    function GetTweetByIndex(const AIndex: Integer): TTwitterTweet;
    function GetTweetByID(const AID: Int64): TTwitterTweet;
    procedure AddTweet(const ATimelineType: TTwitterTimelineType; const ATweet: TTwitterTweet);
  protected
    property UserID: Int64 read FParams.UserID write FParams.UserID default 0;
    property ScreenName: string read FParams.ScreenName write FParams.ScreenName;
    property SinceID: Int64 read FParams.SinceID write FParams.SinceID default 0;
    property Count: Integer read FParams.Count write FParams.Count default 0;
    property MaxID: Int64 read FParams.MaxID write FParams.MaxID default 0;
    property Page: Integer read FParams.Page write FParams.Page default 0;
    property TrimUser: Boolean read FParams.TrimUser write FParams.TrimUser default True;
    property IncludeRetweets: Boolean read FParams.IncludeRetweets write FParams.IncludeRetweets default True;
    property IncludeEntities: Boolean read FParams.IncludeEntities write FParams.IncludeEntities default True;
    property ExcludeReplies: Boolean read FParams.ExcludeReplies write FParams.ExcludeReplies default False;
    property ContributorDetails: Boolean read FParams.ContributorDetails write FParams.ContributorDetails default False;
  public
    constructor Create(const ATimelineManager: TTwitterTimelineManager; const ATimelineType: TTwitterTimelineType);
    destructor Destroy; override;
    procedure Update;
    procedure Clear;
    procedure ClearParams;
    property TweetByIndex[const AIndex: Integer]: TTwitterTweet read GetTweetByIndex; default;
    property TweetByID[const AID: Int64]: TTwitterTweet read GetTweetByID;
  published
    property TweetCount: Integer read FTotal write FTotal default 0;
    property Cache: Boolean read FCache write FCache default True;
    property MaxStore: Integer read FMaxStore write FMaxStore default 0;
  end;

  {
    TTwitterFriendsTimeline is a specialized version of TTwitterTimeline with additional
    properties/methods as required.
    This is deprecated and has been replaced by statuses/home_timeline.
  }
  TTwitterFriendsTimeline = class(TTwitterTimeline)
  published
    property Count;
    property SinceID;
    property MaxID;
    property Page;
    property TrimUser;
    property IncludeRetweets;
    property IncludeEntities;
    property ExcludeReplies;
    property ContributorDetails;
  end;

  {
    TTwitterHomeTimeline is a specialized version of TTwitterTimeline with additional
    properties/methods as required.
  }
  TTwitterHomeTimeline = class(TTwitterTimeline)
  published
    property Count;
    property SinceID;
    property MaxID;
    property Page;
    property TrimUser;
    property IncludeRetweets;
    property IncludeEntities;
    property ExcludeReplies;
    property ContributorDetails;
  end;

  {
    TTwitterMentionsTimeline is a specialized version of TTwitterTimeline with additional
    properties/methods as required.
  }
  TTwitterMentionsTimeline = class(TTwitterTimeline)
  published
    property Count;
    property SinceID;
    property MaxID;
    property Page;
    property TrimUser;
    property IncludeRetweets;
    property IncludeEntities;
    property ContributorDetails;
  end;

  {
    TTwitterPublicTimeline is a specialized version of TTwitterTimeline with additional
    properties/methods as required.
  }
  TTwitterPublicTimeline = class(TTwitterTimeline)
  published
    property TrimUser;
    property IncludeEntities;
  end;

  {
    TTwitterRetweetTimeline is a specialized version of TTwitterTimeline with additional
    properties/methods as required.
  }
  TTwitterRetweetTimeline = class(TTwitterTimeline)
  published
    property Count;
    property SinceID;
    property MaxID;
    property Page;
    property TrimUser;
    property IncludeEntities;
  end;

  {
    TTwitterFriendsTimeline is a specialized version of TTwitterTimeline with additional
    properties/methods as required.
  }
  TTwitterUserTimeline = class(TTwitterTimeline)
  published
    property UserID;
    property ScreenName;
    property SinceID;
    property Count;
    property MaxID;
    property Page;
    property TrimUser;
    property IncludeRetweets;
    property IncludeEntities;
    property ExcludeReplies;
    property ContributorDetails;
  end;

  TTwitterRetweetedTimeline = class(TTwitterTimeline)
  published
    property ScreenName;
    property UserID;
    property Count;
    property SinceID;
    property MaxID;
    property Page;
    property TrimUser;
    property IncludeEntities;
  end;

  {
    TTwitterMultiTimeline object is special as it provides us with a managed array of
    timelines. This is designed to apply to Users managed by Name and/or ID.
  }
  TTwitterMultiTimeline<T: TTwitterTimeline, constructor> = class(TPersistent)
  private
    FCache: Boolean;
    FTimelines: TObjectList<T>;
    FTimelineType: TTwitterTimelineType;
    FTimelineManager: TTwitterTimelineManager;
    function GetTimeline(const AName: string): T;
    function GetTimelineByID(const AID: Int64): T;
  public
    constructor Create(const ATimelineManager: TTwitterTimelineManager; const ATimelineType: TTwitterTimelineType);
    destructor Destroy; override;
    property Timeline[const AName: string]: T read GetTimeline; default;
    property Timeline[const AID: Int64]: T read GetTimelineByID; default;
  published
    property Cache: Boolean read FCache write FCache default True;
  end;

  {
    TTwitterTimelineManager is a special base component providing us a means of
    accessing and managing the various Twitter Timelines.
  }
  TTwitterTimelineManager = class(TPersistent)
  private
    FPublic: TTwitterPublicTimeline;
    FHome: TTwitterHomeTimeline;
    FFriends: TTwitterFriendsTimeline;
    FMentions: TTwitterMentionsTimeline;
    FRetweetedByMe,
    FRetweetedToMe,
    FRetweetsOfMe: TTwitterRetweetTimeline;
    FTwitter: TTwitter;
    FUser: TTwitterMultiTimeline<TTwitterUserTimeline>; // Special! One timeline per user!
    FRetweetedToUser: TTwitterMultiTimeline<TTwitterRetweetedTimeline>;
    FRetweetedByUser: TTwitterMultiTimeline<TTwitterRetweetedTimeline>;
    FNoDuplicates: Boolean;
  public
    constructor Create(const ATwitter: TTwitter); virtual;
    destructor Destroy; override;
  published
    property Public: TTwitterPublicTimeline read FPublic;
    property Home: TTwitterHomeTimeline read FHome;
    property Friends: TTwitterFriendsTimeline read FFriends;
    property Mentions: TTwitterMentionsTimeline read FMentions;
    property RetweetedByMe: TTwitterRetweetTimeline read FRetweetedByMe;
    property RetweetedToMe: TTwitterRetweetTimeline read FRetweetedToMe;
    property RetweetsOfMe: TTwitterRetweetTimeline read FRetweetsOfMe;
    property Users: TTwitterMultiTimeline<TTwitterUserTimeline> read FUser; // Special! One timeline per user!
    property RetweetedToUser: TTwitterMultiTimeline<TTwitterRetweetedTimeline> read FRetweetedToUser;
    property RetweetedByUser: TTwitterMultiTimeline<TTwitterRetweetedTimeline> read FRetweetedByUser;
    property NoDuplicates: Boolean read FNoDuplicates write FNoDuplicates default True;
  end;

const
  TTwitter_Registry = '\Software\%s\Stored\'; // Default Registry location to store Tokens (if no custom instruction given)

implementation

{ TTwitter }

constructor TTwitter.Create;
begin
  Create(nil);
end;

constructor TTwitter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if FApplicationDetails = nil then
    FApplicationDetails := TTwitterApplicationDetails.Create;
  FLoggedIn := False;
  FAutoRateCheck := False;
  FHTTP := TTwitterHTTP.Create(FApplicationDetails);
  FUsers := TTwitterUserManager.Create(Self);
  FProxy := TTwitterProxyDetails.Create;
  FTimelines := TTwitterTimelineManager.Create(Self);
end;

destructor TTwitter.Destroy;
begin
  FreeAndNil(FApplicationDetails);
  if FHTTP <> nil then
    FHTTP.Free;
  FUsers.Free;
  FProxy.Free;
  FTimelines.Free;
  inherited Destroy;
end;

procedure TTwitter.DoStatus(const AResponseCode: Integer; const AMsg: String; const AAction: TTwitterStatusAction; const AStatusType: TTwitterStatusType);
begin
  case AStatusType of
    stError:
      begin
        if Assigned(FOnError) then
          FOnError(AResponseCode, AMsg, AAction)
        else
          raise ETwitterException.Create(AMsg);
      end;
    stException:
      begin
        if Assigned(FOnException) then
          FOnException(AMsg)
        else
          raise ETwitterException.Create(AMsg);
      end;
  end;
end;

function TTwitter.GetDefaultMessage(const AResponseCode: Integer): String;
begin
  case AResponseCode of
    304: Result := TTwitter_Error_304;
    400: Result := TTwitter_Error_400;
    401: Result := TTwitter_Error_401;
    403: Result := TTwitter_Error_403;
    404: Result := TTwitter_Error_404;
    406: Result := TTwitter_Error_406;
    420: Result := TTwitter_Error_420;
    500: Result := TTwitter_Error_500;
    502: Result := TTwitter_Error_502;
    503: Result := TTwitter_Error_503;
  end;
end;

function TTwitter.TwitterGET(const ARequiresAuth: Boolean; var AURL: String): String;
begin
  Result := FHTTP.Get(AURL, ARequiresAuth, False, FProxy);

  if Assigned(FOnRawResponse) then
    FOnRawResponse(AURL, Result);

  if FAutoRateCheck then
    GetRateLimitStatus;
end;

function TTwitter.TwitterPOST(var AURL: String; const AParams: TTwitterParamFactory): String;
begin
  Result := FHTTP.Post(AURL, AParams, False, FProxy);

  if Assigned(FOnRawResponse) then
    FOnRawResponse(AURL, Result);

  if FAutoRateCheck then
    GetRateLimitStatus
end;

{
procedure TTwitter.AsyncTwitterGET(const ARequiresAuth: Boolean; var AURL: String; const ABetterWithAuth: Boolean = True);
begin
  FHTTP.Get(AURL, ARequiresAuth, True, FProxy, ABetterWithAuth);
end;

procedure TTwitter.AsyncTwitterPOST(var AURL: String; const AParams: TTwitterParamFactory);
begin
  FHTTP.Post(AURL, AParams, True, FProxy);
end;
}

procedure TTwitter.GetRateLimitStatus;
var
  URL: String;
  GETResponse: String;
  LRateLimit: TTwitterRateLimitData;
begin
  URL := TTwitterURLAssembler.RateLimitStatus;

  // Get API Limits
  GETResponse := TwitterGET(False, URL);
  LRateLimit := TTwitterParser.ParseRateLimit(GETResponse);

  if Assigned(FOnRateLimitStatus) then
    FOnRateLimitStatus(LRateLimit);
end;

function TTwitter.GetUserInfo(const AUsername: String; const AIncludeEntities: Boolean = True): TTwitterUser;
var
  LURL: String;
  GETResponse: String;
begin
  LURL := TTwitterURLAssembler.UsersShow(AUsername, AIncludeEntities);
  GETResponse := TwitterGET(False, LURL);
  Result := TTwitterParser.ParseUserInfo(GETResponse);
end;

function TTwitter.GetUserInfo(const AID: Int64; const AIncludeEntities: Boolean = True): TTwitterUser;
var
  LURL: String;
  GETResponse: String;
begin
  LURL := TTwitterURLAssembler.UsersShow(AID, AIncludeEntities);
  GETResponse := TwitterGET(False, LURL);
  Result := TTwitterParser.ParseUserInfo(GETResponse);
end;

function TTwitter.Login: Boolean;
begin
  if FLoggedIn then
    Logout;

  FLoggedIn := FHTTP.Login(FUsername, FTokenManager, FProxy);

  Result := FLoggedIn;

  if Result then
  begin
    FUsername := FHTTP.UserToken.Username;
    if Assigned(FOnUsernameChange) then
      FOnUsernameChange(FHTTP.UserToken.Username);
  end;
end;

procedure TTwitter.Logout;
begin
  FHTTP.Logout;
  if FLoggedIn then
    FLoggedIn := False;
end;

procedure TTwitter.ForgetUser;
begin
  if Assigned(FTokenManager) then
    FTokenManager.KillToken(FUserToken);
end;

function TTwitter.PostTweet(const AMsg: String; const AReplyToStatusID: Int64;
  const AImage: string; const ALat, ALong: Extended; const APlaceID: String;
  const ADisplayCoordinates: Boolean; ATrimUser, AIncludeEntities: Boolean): TTwitterTweet;
var
  LParams: TTwitterParamFactory;
  URL,
  GETResponse: String;
begin
  Result := nil;
  LParams := TTwitterParamFactory.Create;
  try
    if (Length(AMsg) = 0) or (Length(AMsg) > 140) then
    begin
      DoStatus(0, TTwitter_Error_StatusLength, saPostTweet, stError);
      Exit;
    end;

    URL := TTwitterURLAssembler.StatusUpdate(AMsg, AReplyToStatusID, AImage, ALat, ALong, APlaceID, ADisplayCoordinates, ATrimUser, AIncludeEntities, LParams);
    try
      GETResponse := TwitterPOST(URL, LParams);
      Result := TTwitterParser.ParseTweet(GETResponse);
    except
      // don't complain that this exception handler somehow "hides" the exception, it doesn't! It simply
      // displays any exceptions using TTwitter's own custom Exception system!
      on E: Exception do
      begin
        if (FHTTP.HTTP.ResponseCode <> 200) and (FHTTP.HTTP.ResponseCode > 300) then
          DoStatus(FHTTP.HTTP.ResponseCode, E.Message, saPostTweet, stException)
        else
          DoStatus(FHTTP.HTTP.ResponseCode, GetDefaultMessage(FHTTP.HTTP.ResponseCode), saPostTweet, stError);
        exit;
      end;
    end;
    if FHTTP.HTTP.ResponseCode <> 200 then
    begin
      // Error Response given by Twitter's API...
      DoStatus(FHTTP.HTTP.ResponseCode, GetDefaultMessage(FHTTP.HTTP.ResponseCode), saPostTweet, stError);
      exit;
    end;
  finally
    LParams.Free;
  end;
end;

function TTwitter.Request(const AMethod: THttpMethod; var AURL: string;
  const AParams: TTwitterParamFactory; const ARequiresAuth: Boolean = True): string;
var
  LParams: string;
begin
  case AMethod of
    GET:
    begin
      LParams := AParams.ToString;
      if Length(LParams) > 0 then
      begin
        if Pos('?', AURL) = 0 then
          AURL := AURL + '?' + LParams
        else
          AURL := AURL + '&' + LParams;
      end;
      Result := TwitterGET(ARequiresAuth, AURL);
    end;
    POST:
      Result := TwitterPOST(AURL, AParams);
  end;
end;

function TTwitter.Request<T>(const AMethod: THttpMethod; var AURL: string;
  const AParams: T; const ARequiresAuth: Boolean): string;
var
  params: TTwitterParamFactory;
begin
  params := TTwitterParamFactory.FromRecord<T>(AParams);
  try
    Result := Request(AMethod, AURL, params, ARequiresAuth);
  finally
    params.Free;
  end;
end;

{
procedure TTwitter.AsyncRequest(const AMethod: THttpMethod; var AURL: string;
  const AParams: TTwitterParamFactory; const ARequiresAuth: Boolean = True);
var
  LParams: string;
begin
  case AMethod of
    GET:
    begin
      LParams := AParams.ToString;
      if Length(LParams) > 0 then
      begin
        if Pos('?', AURL) = 0 then
          AURL := AURL + '?' + LParams
        else
          AURL := AURL + '&' + LParams;
      end;
      AsyncTwitterGET(ARequiresAuth, AURL);
    end;
    POST:
      AsyncTwitterPOST(AURL, AParams);
  end;
end;
}

{ TTwitterUserManager }

constructor TTwitterUserManager.Create(ATwitter: TTwitter);
begin
  inherited Create;
  FCache := True;
  FIncludeEntities := True;
  FTwitter := ATwitter;
end;

destructor TTwitterUserManager.Destroy;
var
  I: Integer;
begin
  for I := 0 to Length(FUsers) - 1 do
  begin
    if FUsers[I] <> nil then
      FreeAndNil(FUsers[I]);
  end;
  SetLength(FUsers, 0);
  inherited;
end;

function TTwitterUserManager.GetUser(const AUsername: String): TTwitterUser;
var
  LIdx: Integer;
begin
  Result := nil;
  if FCache then
  begin
    for LIdx := 0 to Length(FUsers) - 1 do
    begin
      if FUsers[LIdx] <> nil then
      begin
        if FUsers[LIdx].ScreenName =  AUsername then
        begin
          Result := FUsers[LIdx];
          Exit;
        end;
      end;
    end;
  end;
  if Assigned(FTwitter) then
  begin
    LIdx := Length(FUsers);
    SetLength(FUsers, LIdx + 1);
    FUsers[LIdx] := FTwitter.GetUserInfo(AUsername, FIncludeEntities);
    Result := FUsers[LIdx];
  end;
end;

function TTwitterUserManager.GetUserByID(const AID: Int64): TTwitterUser;
var
  LIdx: Integer;
begin
  Result := nil;
  if FCache then
  begin
    for LIdx := 0 to Length(FUsers) - 1 do
    begin
      if FUsers[LIdx] <> nil then
      begin
        if FUsers[LIdx].ID = AID then
        begin
          Result := FUsers[LIdx];
          Exit;
        end;
      end;
    end;
  end;
  if Assigned(FTwitter) then
  begin
    LIdx := Length(FUsers);
    SetLength(FUsers, LIdx + 1);
    FUsers[LIdx] := FTwitter.GetUserInfo(AID, FIncludeEntities);
    Result := FUsers[LIdx];
  end;
end;

{ TTwitterTimeline }

procedure TTwitterTimeline.AddTweet(const ATimelineType: TTwitterTimelineType; const ATweet: TTwitterTweet);
var
  LIdx: Integer;
  LTweetUnique: Boolean;
begin
  if ATweet = nil then Exit;

  if ATweet.ID > FParams.SinceID then
    FParams.SinceId := ATweet.ID;

  LTweetUnique := True;

  if Self.GetTweetByID(ATweet.ID) = nil then
  begin
    LIdx := Length(FTweets);
    SetLength(FTweets, LIdx + 1);
    FTweets[LIdx] := ATweet;
    FTotal := LIdx + 1;
  end else
    LTweetUnique := False;

  if Assigned(FTimelineManager.FTwitter.FOnAsyncTimeline) then
  begin
    if ((not FTimelineManager.FNoDuplicates) and (not LTweetUnique)) or (LTweetUnique)then
      FTimelineManager.FTwitter.FOnAsyncTimeline(ATimelineType, ATweet);
  end;

  if (not LTweetUnique) then
    ATweet.Free;
end;

procedure TTwitterTimeline.Clear;
var
  I: Integer;
begin
  for I := 0 to FTotal -1 do
    FTweets[I].Free;
  FTweets := nil;
end;

procedure TTwitterTimeline.ClearParams;
begin
  with FParams do
  begin
    UserID := 0;
    ScreenName := EmptyStr;
    SinceID := 0;
    Count := 0;
    MaxID := 0;
    Page := 0;
    TrimUser := False;
    IncludeRetweets := True;
    IncludeEntities := True;
    ExcludeReplies := False;
    ContributorDetails := False;
  end;
end;

constructor TTwitterTimeline.Create(const ATimelineManager: TTwitterTimelineManager; const ATimelineType: TTwitterTimelineType);
begin
  FTimelineManager := ATimelineManager;
  FTimelineType := ATimelineType;
  FCache := True;
  FMaxStore := 0;
  FTotal := 0;
  ClearParams;
end;

destructor TTwitterTimeline.Destroy;
begin
  Clear;
  inherited;
end;

procedure TTwitterTimeline.Update;
var
  LURL, GETResponse: String;
begin
  if not Assigned(FTimelineManager.FTwitter) then Exit;
  if FTimelineType = ttNone then Exit; // TODO: Exception Handler

  LURL := TTwitterURLAssembler.Timeline(FTimelineType, FParams);

  GETResponse := FTimelineManager.FTwitter.TwitterGET((FTimelineType <> ttUser) and (FTimelineType <> ttPublic), LURL);
  // TODO: Exception Handler

  TTwitterParser.ParseTimeline(GETResponse, AddTweet, FTimelineType);
end;

function TTwitterTimeline.GetTweetByID(const AID: Int64): TTwitterTweet;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to FTotal - 1 do
  begin
    if FTweets[I].ID = AID then
    begin
      Result := FTweets[I];
      Break;
    end;
  end;
end;

function TTwitterTimeline.GetTweetByIndex(const AIndex: Integer): TTwitterTweet;
begin
  if (AIndex >= 0) and (AIndex <= FTotal) then
    Result := FTweets[AIndex]
  else
    Result := nil;
end;

{ TTwitterMultiTimeline<T> }

constructor TTwitterMultiTimeline<T>.Create(const ATimelineManager: TTwitterTimelineManager; const ATimelineType: TTwitterTimelineType);
begin
  inherited Create;
  FTimelineManager := ATimelineManager;
  FTimelineType := ATimelineType;
  FTimelines := TObjectList<T>.Create;
end;

destructor TTwitterMultiTimeline<T>.Destroy;
begin
  FTimelines.Free;
  inherited;
end;

function TTwitterMultiTimeline<T>.GetTimeline(const AName: string): T;
begin
  for Result in FTimelines do
    if SameText(Result.ScreenName, AName) then Exit;

  Result := T.Create;
  Result.Create(FTimelineManager, FTimelineType);
  Result.ScreenName := AName;
  FTimelines.Add(Result);
end;

function TTwitterMultiTimeline<T>.GetTimelineByID(const AID: Int64): T;
begin
  for Result in FTimelines do
    if Result.UserID = AID then Exit;

  Result := T.Create;
  Result.Create(FTimelineManager, FTimelineType);
  Result.UserID := AID;
  FTimelines.Add(Result);
end;

{ TTwitterTimelineManager }

constructor TTwitterTimelineManager.Create(const ATwitter: TTwitter);
begin
  inherited Create;
  FPublic := TTwitterPublicTimeline.Create(Self, ttPublic);
  FHome := TTwitterHomeTimeline.Create(Self, ttHome);
  FFriends := TTwitterFriendsTimeline.Create(Self, ttFriends);
  FMentions := TTwitterMentionsTimeline.Create(Self, ttMentions);
  FRetweetedByMe := TTwitterRetweetTimeline.Create(Self, ttRetweetedByMe);
  FRetweetedToMe := TTwitterRetweetTimeline.Create(Self, ttRetweetedToMe);
  FRetweetsOfMe := TTwitterRetweetTimeline.Create(Self, ttRetweetsOfMe);
  FUser := TTwitterMultiTimeline<TTwitterUserTimeline>.Create(Self, ttUser);
  FRetweetedToUser := TTwitterMultiTimeline<TTwitterRetweetedTimeline>.Create(Self, ttRetweetedToUser);
  FRetweetedByUser := TTwitterMultiTimeline<TTwitterRetweetedTimeline>.Create(Self, ttRetweetedByUser);
  FTwitter := ATwitter;
  FNoDuplicates := True;
end;

destructor TTwitterTimelineManager.Destroy;
begin
  FPublic.Free;
  FHome.Free;
  FFriends.Free;
  FMentions.Free;
  FRetweetedByMe.Free;
  FRetweetedToMe.Free;
  FRetweetsOfMe.Free;
  FUser.Free;
  FRetweetedToUser.Free;
  FRetweetedByUser.Free;
  inherited;
end;

end.
