{
  Original Name: Twitter_HTTP.pas
  Copyright: Copyright (c) 2010, Simon J Stuart, All Rights Reserved
  License: Licensed under the MIT License.
  Author: Simon J Stuart
  Author's Website: http://www.lakraven.com
  Description:
    This unit contains everything related to dealing with HTTP requests.
    It incorporates the OAuth routines, all of which have been abstracted
    to make the code easier to debug and read.
  Author's Notes: (See "Twitter.pas")
  Acknowledgements: (See "Twitter.pas")

  Version: 0.1 OFFICIAL OPEN BETA RELEASE
  --------------------------------------------------------------------------------
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
  --------------------------------------------------------------------------------
  CODE NOTES: (See "Twitter.pas")

  IMPORTANT NOTICE: You are currently looking at an Open BETA version of TTwitter!
                    This code is subject to a LOT of changes, and whilst you may
                    find portions of this code at present to be messy... rest
                    assured that upon final release, it'll all be nice and clean!
}
unit Twitter_HTTP;

interface

{$I Twitter.inc}

uses
  Windows, Classes, SysUtils, Forms, StrUtils,
{$IFDEF INDY10}
  IdBaseComponent, IdComponent, IdTCPConnection, IdTCPClient, IdHTTP, IdURI, IdSSLOpenSSLHeaders,
  IdIOHandler, IdIOHandlerSocket, IdIOHandlerStack, IdSSL, IdSSLOpenSSL, IdMultipartFormData,
  IdIntercept, IdGlobal,
{$ELSE}
  {$IFDEF INDY9}
    ERROR: Older versions are not yet supported ! You can always install Indy 10 on Older Delphi versions!
           {Get Indy 10 here: http://www.indyproject.org/Sockets/Download/Files/Indy10.EN.aspx}
  {$ELSE}
    ERROR: You must have at least ONE version of Indy installed! Get Indy 10 from the above URL!
  {$ENDIF}
{$ENDIF}
  Twitter_Common, Twitter_OAuth, Twitter_OLogin, Twitter_StringComposer, Twitter_TokenStore;

type

  {
    TTwitterHTTP object deals with INDY's TIdHTTP component. I've abstracted this away from
    the main Twitter.pas unit to make the code easier to read and follow, and because
    the more abstract a system is, the easier it becomes to maintain each subsystem.
  }
  TTwitterHTTP = class(TObject)
  private
    FAppDetails: TTwitterApplicationDetails;
    FHTTP: TIdHTTP;
    FUserToken: TTwitterUserToken;
    FSSL: TIdSSLIOHandlerSocketOpenSSL;
    FInnerBuffer: UTF8String;
    function UserAuthorization(const AURL, AUsername: string): string;
    procedure InterceptReceive(ASender: TIdConnectionIntercept; var ABuffer: TIdBytes);
  public
    constructor Create(const AAppDetails: TTwitterApplicationDetails);
    destructor Destroy; override;
    procedure New(const AURL: string; const AAsync: Boolean; const AProxy: TTwitterProxyDetails);
    function Login(var AUsername: String; const ATokenManager: TTwitterTokenManager; const AProxy: TTwitterProxyDetails): Boolean;
    procedure Logout;
    function Get(var AURL: string; const ARequiresAuth: Boolean; const AAsync: Boolean;
      const AProxy: TTwitterProxyDetails; const Verifier: string = ''): string;
    function Post(var AURL: string; const AParams: TTwitterParamFactory;
      const AAsync: Boolean; const AProxy: TTwitterProxyDetails): string;
    property HTTP: TIdHTTP read FHTTP;
    property UserToken: TTwitterUserToken read FUserToken;
  end;

implementation

{ TTwitterHTTP }

function TTwitterHTTP.UserAuthorization(const AURL, AUsername: string): string;
var
  LTwitterLogin: TTwitterLogin;
begin
  Application.CreateForm(TTwitterLogin, LTwitterLogin);
  try
    with LTwitterLogin do
    begin
      UserName := AUsername;
      ValidURL := AURL;
      TwitterLoginPage.Navigate(AURL);
      ShowModal;
      Result := PIN;
    end;
  finally
    LTwitterLogin.Free;
  end;
end;

constructor TTwitterHTTP.Create(const AAppDetails: TTwitterApplicationDetails);
begin
  inherited Create;
  FAppDetails := AAppDetails
end;

destructor TTwitterHTTP.Destroy;
begin
  if FHTTP <> nil then
    FreeAndNil(FHTTP);
  if FSSL <> nil then
    FreeAndNil(FSSL);
  inherited;
end;

function TTwitterHTTP.Login(var AUsername: String; const ATokenManager: TTwitterTokenManager; const AProxy: TTwitterProxyDetails): Boolean;
const
  TWITTER_OAUTH_TOKENPARAM = 'oauth_token';
  TWITTER_OAUTH_SECRETPARAM = 'oauth_token_secret';
  TWITTER_OAUTH_NAMEPARAM = 'screen_name';
  TWITTER_OAUTH_IDPARAM = 'user_id';
  TWITTER_OAUTH_AUTHORIZE = '%s?%s=%s&%s=%s'; // <url>?oauth_token=<value>&oauth_token_secret=<value>
var
  LStringSplit: TTwitterStringSplitter;
  LURL,
  LGETResponse: String;
  LPin: string;
begin
  // ENSURE THAT ALL NECESSARY VALUES HAVE BEEN PROVIDED
  with FAppDetails do
  begin
    Assert(ApiKey <> EmptyStr, TTwitter_Except_MissingAPIKey);
    Assert(ApiSecret <> EmptyStr, TTwitter_Except_MissingAPISecret);
  end;

  // Let's see if we can retrieve a Stored Token for this User!
  if (Assigned(ATokenManager)) and (AUsername <> EmptyStr) then
  begin
    FUserToken.Username := AUsername; // Pass our Username into the UserToken object
    ATokenManager.GetToken(FUserToken); // Pass our UserToken object to the Token Manager to request that the Token and Secret be populated
    if (FUserToken.Token <> EmptyStr) and (FUserToken.TokenSecret <> EmptyStr) and (FUserToken.Username <> EmptyStr) then
    begin
      // We have been given our User Token values, let's validate them with Twitter.
      { TODO -oLaKraven -cOAuth - Token Validation : Figure out how to Authenticate a User OAuth Token with Twitter's API (Documentation approach failed completely in testing) }
      Result := True;
      Exit; // We can exit here because having a valid User Token means we are already logged in!
    end;
  end;

  { Everything else in this function only occurs where no Stored User Token can be found! }

  LStringSplit := TTwitterStringSplitter.Create;
  try
    LURL := TTwitterURLAssembler.OAuth(oaRequestToken);
    LGETResponse := Get(LURL, True, False, AProxy); // This should give us an Auth Token (NOT the same as a User Token)
    LStringSplit.ParseURLParams(LGETResponse); // This works some magic to split the "URL Param String" response into Key:Value pairs!

    with FUserToken do // We temporarily store our Auth Token in the User Token container (NOTE: Auth Tokens and User Tokens are totally different)
    begin
      Token := LStringSplit[TWITTER_OAUTH_TOKENPARAM];
      TokenSecret := LStringSplit[TWITTER_OAUTH_SECRETPARAM];
    end;

    // Now we prepare our User Authorization URL String from our Auth Token (NOT a User Token)
    LURL := Format(TWITTER_OAUTH_AUTHORIZE, [TTwitterURLAssembler.OAuth(oaAuthorize), TWITTER_OAUTH_TOKENPARAM, FUserToken.Token, TWITTER_OAUTH_SECRETPARAM, FUserToken.TokenSecret]);
    LPin := UserAuthorization(LURL, AUsername); // We are now ready to display our Twitter Authorization Form

    if LPin = EmptyStr then
    begin
      Result := False;
      Exit;
    end;

    // Hopefully at this point the User will have selected "Allow" so we can proceed to retrieve their User Token!
    LURL := TTwitterURLAssembler.OAuth(oaAccessToken);
    LGETResponse := Get(LURL, True, False, AProxy, LPin); // Retrieve the User Token
    LStringSplit.ParseURLParams(LGETResponse);
    with FUserToken do // We now store our actual User Token details
    begin
      Token := LStringSplit[TWITTER_OAUTH_TOKENPARAM];
      TokenSecret := LStringSplit[TWITTER_OAUTH_SECRETPARAM];
      Username := LStringSplit[TWITTER_OAUTH_NAMEPARAM];
      UserID := StrToInt64(LStringSplit[TWITTER_OAUTH_IDPARAM]);
      AUsername := Username;
    end;

    if (Assigned(ATokenManager)) and (FUserToken.Username <> EmptyStr) then // Store our User Token (if a Token Manager is available)
      ATokenManager.StoreToken(FUserToken);

    Result := True;
  finally
    LStringSplit.Free;
  end;
end;

procedure TTwitterHTTP.Logout;
begin
  with FUserToken do
  begin
    Token := EmptyStr;
    TokenSecret := EmptyStr;
    Username := EmptyStr;
    UserID := 0;
  end;
end;

procedure TTwitterHTTP.InterceptReceive(ASender: TIdConnectionIntercept; var ABuffer: TIdBytes);
var
  i: Integer;
  s: string;
begin
  i := Length(FInnerBuffer) + 1;
  SetLength(FInnerBuffer, Length(FInnerBuffer)+Length(ABuffer));
  Move(ABuffer[0], FInnerBuffer[i], Length(ABuffer));
  while True do
  begin
    i := Pos(UTF8String(#13#10), FInnerBuffer);
    if i = 0 then Break;
    if i = 1 then
    begin
      FInnerBuffer := Copy(FInnerBuffer, 3, Length(FInnerBuffer));
      Continue;
    end;
    s := string(Copy(FInnerBuffer, 1, i-1));
    FInnerBuffer := Copy(FInnerBuffer, i+2, Length(FInnerBuffer));
    if (Length(s) > 0) and (s[1] = '{') then
      OutputDebugString(PChar('[Streaming]'+s));
  end;
end;

procedure TTwitterHTTP.New(const AURL: string; const AAsync: Boolean; const AProxy: TTwitterProxyDetails);
begin
  if FSSL <> nil then
    FreeAndNil(FSSL);
  if FHTTP <> nil then
    FreeAndNil(FHTTP);

  FHTTP := TIdHTTP.Create(nil);
  with FHTTP do
  begin
    if AnsiPos('https:', AURL) > 0 then
    begin
      FSSL := TIdSSLIOHandlerSocketOpenSSL.Create(FHTTP);
      IOHandler := FSSL;
    end;
    Request.UserAgent := 'Mozilla/3.0 (compatible; TTwitter)';
    Request.Accept := 'text/html, */*';
    HTTPOptions := [hoForceEncodeParams, hoInProcessAuth];
    MaxAuthRetries := 3;
    Request.ContentType := 'application/x-www-form-urlencoded';
    ConnectTimeout := 9000;
    HTTPOptions := [];
    with ProxyParams do
    begin
      BasicAuthentication := AProxy.BasicAuthentication;
      ProxyPassword := AProxy.Password;
      ProxyPort := AProxy.Port;
      ProxyServer := AProxy.Server;
      ProxyUsername := AProxy.Username;
    end;
    if AAsync then
    begin
      Intercept := TIdConnectionIntercept.Create(FHTTP);
      Intercept.OnReceive := InterceptReceive;
    end;
  end;
end;

function TTwitterHTTP.Get(var AURL: string; const ARequiresAuth: Boolean;
  const AAsync: Boolean; const AProxy: TTwitterProxyDetails; const Verifier: string): string;
var
  LConsumer: TOAuthConsumer;
  LRequest: TOAuthRequest;
  LToken: TOAuthToken;
  LParams: TStringList;
  LURL: string;
  LAuthHeader: string;
  StringSplitter: TTwitterStringSplitter;
  I: Integer;
begin
{ Lines below commented due to being incomplete }
//  if (ARequiresAuth) and ((FUserToken.Token = EmptyStr) or (FUserToken.TokenSecret = EmptyStr) or (FUserToken.Username = EmptyStr)) then begin
//    if (not FLoginMethod) then
      { TODO -oLaKraven -cTTwitter Error Handler : Report back to the user that we're not logged in }
//  end;

  New(AURL, AAsync, AProxy); // Creates a clean IdHTTP instance to avoid ambiguity!
  LURL := AURL; // Just in case we have an SSL failure, we can reconstruct from the original URL
  try
    if ARequiresAuth then
    begin
      LConsumer := TOAuthConsumer.Create(FAppDetails.ApiKey, FAppDetails.ApiSecret);
      LRequest := TOAuthRequest.Create(AURL);
      LToken := TOAuthToken.Create(FUserToken.Token, FUserToken.TokenSecret);
      LRequest := LRequest.FromConsumerAndToken(LConsumer, LToken, AURL, Verifier);
      LRequest.HTTPURL := AURL;

      LParams := TStringList.Create;
      try
        StringSplitter := TTwitterStringSplitter.Create;
        try
          StringSplitter.OrderURL(AURL);
          for I := Low(StringSplitter.Params) to High(StringSplitter.Params) do
            LParams.Add(StringSplitter.Params[I].Key+'='+StringSplitter.Params[I].Value);

          I := Pos('?', AURL);
          if I  > 0 then
            LAuthHeader := LRequest.genAuthHeader('GET', LConsumer, LToken, LParams, Copy(AURL, 1, I-1))
          else
            LAuthHeader := LRequest.genAuthHeader('GET', LConsumer, LToken, LParams, AURL);
          FHTTP.Request.CustomHeaders.Add('Authorization: ' + LAuthHeader);
        finally
          StringSplitter.Free;
        end;
      finally
        LParams.Free;
      end;
    end;
    try
      Result := FHTTP.Get(AURL);
    except
      on EIdOSSLCouldNotLoadSSLLibrary do
      begin
        AURL := ReplaceText(LURL, 'https://', 'http://');
        Result := Get(AURL, ARequiresAuth, AAsync, AProxy);
      end;
    end;
  finally
    if ARequiresAuth then
    begin
      FreeAndNil(LConsumer);
      FreeAndNil(LRequest);
      FreeAndNil(LToken);
    end;
  end;
end;

function TTwitterHTTP.Post(var AURL: string; const AParams: TTwitterParamFactory;
  const AAsync: Boolean; const AProxy: TTwitterProxyDetails): string;
var
  LConsumer: TOAuthConsumer;
  LRequest: TOAuthRequest;
  LToken: TOAuthToken;
  LParams: TStringList;
  LMultiPart: TIdMultiPartFormDataStream;
  I: Integer;
  LAuthHeader: String;
begin
  if ((FUserToken.Token = EmptyStr) or (FUserToken.TokenSecret = EmptyStr) or (FUserToken.Username = EmptyStr)) then
  begin
    { TODO -oLaKraven -cTTwitter Error Handler : Report back to the user that we're not logged in }
    Exit;
  end;

  New(AURL, AAsync, AProxy); // Creates a clean IdHTTP instance to avoid ambiguity!

  if AParams.IsMultiPart then
  begin
    try
      LConsumer := TOAuthConsumer.Create(FAppDetails.ApiKey, FAppDetails.ApiSecret);
      LToken := TOAuthToken.Create(FUserToken.Token, FUserToken.TokenSecret);
      LRequest := TOAuthRequest.Create(AURL);
      LRequest := LRequest.FromConsumerAndToken(LConsumer, LToken, AURL);
      LRequest.HTTPURL := AURL;
      LAuthHeader := LRequest.genAuthHeader('POST', LConsumer, LToken, nil, AURL);
      FHTTP.Request.CustomHeaders.Add('Authorization: ' + LAuthHeader);

      LMultiPart := TIdMultiPartFormDataStream.Create;
      try
        for I := 0 to AParams.Count-1 do
        begin
          if AParams[i].IsFile then
            LMultiPart.AddFile(AParams[I].Key, AParams[I].Value, '')
          else
            LMultiPart.AddFormField(AParams[I].Key, AParams[I].Value, 'UTF-8')
              .ContentTransfer := sContentTransferBinary;
        end;
        Result := FHTTP.Post(AURL, LMultiPart);
      finally
        LMultiPart.Free;
      end;
    finally
      FreeAndNil(LConsumer);
      FreeAndNil(LRequest);
      FreeAndNil(LToken);
    end
  end
  else begin
    LParams := TStringList.Create;
    try
      for I := 0 to AParams.Count - 1 do
        LParams.Add(AParams[I].ToString);
      try
        LConsumer := TOAuthConsumer.Create(FAppDetails.ApiKey, FAppDetails.ApiSecret);
        LToken := TOAuthToken.Create(FUserToken.Token, FUserToken.TokenSecret);
        LRequest := TOAuthRequest.Create(AURL);
        LRequest := LRequest.FromConsumerAndToken(LConsumer, LToken, AURL);
        LRequest.HTTPURL := AURL;
        LAuthHeader := LRequest.genAuthHeader('POST', LConsumer, LToken, LParams, AURL);
        FHTTP.Request.CustomHeaders.Add('Authorization: ' + LAuthHeader);
        LParams.Text := LRequest.encodeParams(LParams, '&', False, {$IFDEF SUPPORTS_UNICODE}True{$ELSE}False{$ENDIF});
        Result := FHTTP.Post(AURL, LParams);
      finally
        FreeAndNil(LConsumer);
        FreeAndNil(LRequest);
        FreeAndNil(LToken);
      end;
    finally
      LParams.Free;
    end;
  end;
end;

end.
